local oric = 1 -- 1 for oric constraints

local width,height = 240,200
local w39 = math.floor(width/6)-1
local running = 1

Color = {}
function Color:new(r,g,b)
	local o = {};
	o.r = type(r)=='number' and r or r and r.r or 0;
	o.g = type(g)=='number' and g or r and r.g or 0;
	o.b = type(b)=='number' and b or r and r.b or 0;
	setmetatable(o, self)
	self.__index = self
	return o
end
Color.black = Color:new()
Color.white = Color:new(Color.ONE,Color.ONE,Color.ONE)
function Color:tostring()
	return "(r=" .. self.r .. " g=" .. self.g .. " b=" .. self.b .. ")"
end
function Color:map(func)
	self.r = func(self.r)
	self.g = func(self.g)
	self.b = func(self.b)
	return self
end

function Color:mul(val) return self:map(function(x) return val*x end) end
function Color:div(val) return self:mul(1/val) end
function Color:add(other, coef)
	if coef then
		self.r = self.r + other.r*coef;
		self.g = self.g + other.g*coef;
		self.b = self.b + other.b*coef;
	else
		self.r = self.r + other.r;
		self.g = self.g + other.g;
		self.b = self.b + other.b;
	end
	return self
end
function Color:sub(other, coef) return self:add(other, coef and -coef or -1) end	
function Color:toLinear() 
	return self:map(function(val) 
			val = val/255
			-- if val<=0.081 then val = val/4.5; else val = ((val+0.099)/1.099)^2.2; end
			
			-- works much metter: https://fr.wikipedia.org/wiki/SRGB#Transformation_inverse
			if val<=0.04045 then val = val/12.92 else val = ((val+0.055)/1.055)^2.4 end
			return val
		end)
end
function Color:toPC() 
	return self:map(function(val) 
			-- if val<=0.018 then val = 4.5*val; else val = 1.099*(val ^ (1/2.2))-0.099; end
			-- works much metter: https://fr.wikipedia.org/wiki/SRGB 
			if val<=0.0031308 then val=12.92*val else val = 1.055*(val ^ (1/2.4))-0.055 end
			return val*255
		end)
end

-- return the Color @(x,y) on the original screen in linear space
local screen_w, screen_h, _getLinearPictureColor = getpicturesize()
function getLinearPictureColor(x,y) 
	if _getLinearPictureColor==nil then
		_getLinearPictureColor = {}
		for i=0,255 do _getLinearPictureColor[i] = Color:new(getbackupcolor(i)):toLinear(); end
	end
	return (x<0 or y<0 or x>=screen_w or y>=screen_h) and Color.black or _getLinearPictureColor[getbackuppixel(x,y)]
end
	
local function dist2(c1,c2)
	-- do return (c1.r - c2.r)^2 + (c1.g - c2.g)^2 + (c1.b - c2.b)^2 end
	
	if true then
		-- http://www.compuphase.com/cmetric.htm#GAMMA
		local rM = (c1.r+c2.r)*.5		
		return ((c1.r-c2.r)^2)*(2+rM) +
			   ((c1.g-c2.g)^2)*(4+1) +
			   ((c1.b-c2.b)^2)*(3-rM)
	end

	local x = math.max(c1.r,c2.r)
	local y = math.max(c1.g,c2.g)
	local z = math.max(c1.b,c2.b)
	local u = 4
	return (u+x)*(c1.r - c2.r)^2 + (u+y)*(c1.g - c2.g)^2 + (u+z)*(c1.b - c2.b)^2
end

function info(...)
	local txt = ""
	for _,t in ipairs({...}) do txt = txt .. t end
	statusmessage(txt); 
	if waitbreak(0)==1 then 
		local ok=false
		selectbox("Abort ?", "Yes", function() ok = true end, "No", function() ok = false end)
		if ok then
			running = 0
			-- error('Operation aborted') 
		end
	end
	
end

function to_screen(x,y)
	local i,j;
	if screen_w/screen_h < width/height then
		i = x*screen_h/height
		j = y*screen_h/height
		i = i + (screen_w - width*screen_h/height)/2 -- center
	else
		i = x*screen_w/width
		j = y*screen_w/width
		-- j = j + (screen_h - height*screen_w/width)/2 -- center
	end
	return math.floor(i), math.floor(j)
end

-- return the Color @(x,y) in linear space (0-255)
-- corresonding to the other platform screen
local _getLinearPixel = {} -- cache
function getLinearPixel(x,y)
	if x<6 then return Color:new() end
	local k=x+y*width
	local p = _getLinearPixel and _getLinearPixel[k]
	if not p then
		local x1,y1 = to_screen(x,y)
		local x2,y2 = to_screen(x+1,y+1)
		if x2==x1 then x2=x1+1 end
		if y2==y1 then y2=y1+1 end

		p = Color:new(0,0,0)
		for j=y1,y2-1 do
			for i=x1,x2-1 do
				p:add(getLinearPictureColor(i,j))
			end
		end
		p:div((y2-y1)*(x2-x1)) -- :floor()
		-- p:mul(.95)
		-- local function z() return (math.random()+math.random()-1) end
		-- p.r,p.g,p.b = Color.clamp(p.r + z(), p.g + z(), p.b + z())
					
		if _getLinearPixel then 
			_getLinearPixel[k]=p 
		end
	end
	
	return _getLinearPixel and Color:new(p) or p
end

function mkOstro()
	local o = {}
	local t = { 
		13,     0,     5,
		13,     0,     5,
		21,     0,    10,
		 7,     0,     4,
		 8,     0,     5,
		47,     3,    28,
		23,     3,    13,
		15,     3,     8,
		22,     6,    11,
		43,    15,    20,
		 7,     3,     3,
	   501,   224,   211,
	   249,   116,   103,
	   165,    80,    67,
	   123,    62,    49,
	   489,   256,   191,
		81,    44,    31,
	   483,   272,   181,
		60,    35,    22,
		53,    32,    19,
	   237,   148,    83,
	   471,   304,   161,
		 3,     2,     1,
	   459,   304,   161,
		38,    25,    14,
	   453,   296,   175,
	   225,   146,    91,
	   149,    96,    63,
	   111,    71,    49,
		63,    40,    29,
		73,    46,    35,
	   435,   272,   217,
	   108,    67,    56,
		13,     8,     7,
	   213,   130,   119,
	   423,   256,   245,
		 5,     3,     3,
	   281,   173,   162,
	   141,    89,    78,
	   283,   183,   150,
		71,    47,    36,
	   285,   193,   138,
		13,     9,     6,
		41,    29,    18,
		36,    26,    15,
	   289,   213,   114,
	   145,   109,    54,
	   291,   223,   102,
		73,    57,    24,
	   293,   233,    90,
		21,    17,     6,
	   295,   243,    78,
		37,    31,     9,
		27,    23,     6,
	   149,   129,    30,
	   299,   263,    54,
		75,    67,    12,
		43,    39,     6,
	   151,   139,    18,
	   303,   283,    30,
		38,    36,     3,
	   305,   293,    18,
	   153,   149,     6,
	   307,   303,     6,
		 1,     1,     0, -- 64
	   101,   105,     2,
		49,    53,     2,
		95,   107,     6,
		23,    27,     2,
		89,   109,    10,
		43,    55,     6,
		83,   111,    14,
		 5,     7,     1,
	   172,   181,    37,
		97,    76,    22,
		72,    41,    17,
	   119,    47,    29,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		65,    18,    17,
		95,    29,    26,
	   185,    62,    53,
		30,    11,     9,
		35,    14,    11,
		85,    37,    28,
		55,    26,    19,
		80,    41,    29,
	   155,    86,    59,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
	   305,   176,   119,
	   155,    86,    59,
	   105,    56,    39,
		80,    41,    29,
		65,    32,    23,
		55,    26,    19,
	   335,   152,   113,
		85,    37,    28,
	   115,    48,    37,
		35,    14,    11,
	   355,   136,   109,
		30,    11,     9,
	   365,   128,   107,
	   185,    62,    53,
		25,     8,     7,
		95,    29,    26,
	   385,   112,   103,
		65,    18,    17,
	   395,   104,   101,
		 4,     1,     1,
		 4,     1,     1,
	   395,   104,   101,
		65,    18,    17,
	   385,   112,   103,
		95,    29,    26,
		25,     8,     7,
	   185,    62,    53,
	   365,   128,   107,
		30,    11,     9,
	   355,   136,   109,
		35,    14,    11,
	   115,    48,    37,
		85,    37,    28,
	   335,   152,   113,
		55,    26,    19,
		65,    32,    23,
		80,    41,    29,
	   105,    56,    39,
	   155,    86,    59,
	   305,   176,   119,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
		 5,     3,     2,
	   155,    86,    59,
		80,    41,    29,
		55,    26,    19,
		85,    37,    28,
		35,    14,    11,
		30,    11,     9,
	   185,    62,    53,
		95,    29,    26,
		65,    18,    17,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
		 4,     1,     1,
	   119,    47,    29,
		72,    41,    17,
		97,    76,    22,
	   172,   181,    37,
		 5,     7,     1,
		83,   111,    14,
		43,    55,     6,
		89,   109,    10,
		23,    27,     2,
		95,   107,     6,
		49,    53,     2,
	   101,   105,     2,
		 1,     1,     0,
	   307,   303,     6,
	   153,   149,     6,
	   305,   293,    18,
		38,    36,     3,
	   303,   283,    30,
	   151,   139,    18,
		43,    39,     6,
		75,    67,    12,
	   299,   263,    54,
	   149,   129,    30,
		27,    23,     6,
		37,    31,     9,
	   295,   243,    78,
		21,    17,     6,
	   293,   233,    90,
		73,    57,    24,
	   291,   223,   102,
	   145,   109,    54,
	   289,   213,   114,
		36,    26,    15,
		41,    29,    18,
		13,     9,     6,
	   285,   193,   138,
		71,    47,    36,
	   283,   183,   150,
	   141,    89,    78,
	   281,   173,   162,
		 5,     3,     3,
	   423,   256,   245,
	   213,   130,   119,
		13,     8,     7,
	   108,    67,    56,
	   435,   272,   217,
		73,    46,    35,
		63,    40,    29,
	   111,    71,    49,
	   149,    96,    63,
	   225,   146,    91,
	   453,   296,   175,
		38,    25,    14,
	   459,   304,   161,
		 3,     2,     1,
	   471,   304,   161,
	   237,   148,    83,
		53,    32,    19,
		60,    35,    22,
	   483,   272,   181,
		81,    44,    31,
	   489,   256,   191,
	   123,    62,    49,
	   165,    80,    67,
	   249,   116,   103,
	   501,   224,   211,
		 7,     3,     3,
		43,    15,    20,
		22,     6,    11,
		15,     3,     8,
		23,     3,    13,
		47,     3,    28,
		 8,     0,     5,
		 7,     0,     4,
		21,     0,    10,
		13,     0,     5,
		13,     0,     5}
	for i=1,#t,3 do
		local a,b,c = t[i],t[i+1],t[i+2]
		local d = (a+b+c) -- *(1+0.1*(i/768)) -- *ATT
		table.insert(o, {a/d, b/d, c/d})
	end
	o.diffuse = function(self, col, err, err0, err1, err2)
		-- pas mal
		local a,b = math.max(col.r, col.g, col.b),
				    math.min(col.r, col.g, col.b)
		local z = a==0 and 0 or (a-b)/(a+b)
		-- z = 0.9992 - 0.003*math.exp(1-1/z-1/a)*1.9
		-- z = 0.9981 - 0.0176*math.exp(-1/z-1/a)
		-- z = 0.998 - 0.003*math.exp(1-1/z)*1.85*a^2
		z = 0.9974 - math.exp(-1/z)*0.0152 -- *a
		
		-- local a = math.max(col.r, col.g, col.b)
		-- local z = a==0 and 0 or (((col.r/a - .5)^2 + (col.g/a - .5)^2 + (col.b/a -.5)^2)/.75)^.5
		-- z = 0.9992 - 1.81*0.0241*math.exp(1-1/z)*(a/Color.ONE)^2 -- okayyyy
		-- z = 0.9992 - 0.0241 * math.exp(2-1/z-Color.ONE/a)*0
		
		-- z = 0.9992  - 0.053*math.exp(1-1/z)*a^2
		-- z = 0.999 - 0.05*math.exp(1-1/z-1/a)
		
		-- z = 0.999
		local function d(rgb)
			local e = err[rgb]*z
			local c = self[math.floor(1.5 + 255*col[rgb])]
			if err0 then
				err0[rgb] = err0[rgb] + e*c[1]
				err1[rgb] = err1[rgb] + e*c[2]
				err2[rgb] = err2[rgb] + e*c[3]
			else
				err[rgb] = e*c[1]
			end
		end
		d("r"); d("g"); d("b")
	end
	
	return o
end

local ostro = mkOstro()

function mkPal() 
	local t = {}
	local function btst(val, bit)
		return math.floor(val/bit) % 2 > 0
	end
	for i=0,7 do
		t[i] = Color:new(
				 btst(i,1) and 1 or 0,
				 btst(i,2) and 1 or 0,
				 btst(i,4) and 1 or 0)
	end
	t.disp = function(self)
		for i=0,7 do
			setcolor(i, math.floor(self[i].r*255), math.floor(self[i].g*255), math.floor(self[i].b*255))
		end
	end
	return t
end
local pal = mkPal()

function mkErr(w)
	local e = {}
	for i=-1,w do e[i] = Color:new(0,0,0) end
	e.clear = function(self) 
		for i=-1,w do self[i]:mul(0) end
	end
	return e
end
local err1,err2 = mkErr(width),mkErr(width)

function mkLine(y,err1)
	local line={y=y,err1=err1,e = Color:new()}
		
	for i=0,w39 do
		line[i] = {
			fg=7,
			bg=0,
			cmd=64, 
			i=i,
			cache = {},
			calcErr = function(self,fg,bg)
				local k = bg*8 + fg + 64
				local t = self.cache[k]
				if t==nil then
					local l = fg*8 + bg + 64
					t,fg,bg = 0,pal[fg],pal[bg]
					local e = Color:new()
					-- local A,B = Color:new(),Color:new()
					for x=self.i*6,self.i*6+5 do
						local p = getLinearPixel(x,line.y)
						-- A:add(p):add(line.err1[x])
						e:add(p):add(line.err1[x])
						local c = fg
						local d = dist2(e, c)
						if bg~=fg then
							local t = dist2(e, bg)
							if t<d then c,d = bg,t end
						end
						t = t + d
						-- B:add(c)
						e:sub(c)
						ostro:diffuse(p,e)
					end
					-- t = t + dist2(A:div(6),B:div(6))*4
					self.cache[k] = t
					self.cache[l] = t
				end
				return t
			end
		}
	end
	
	function line:findRec(x, ink, pap)
		-- done ?
		if x>w39 then return 0 end

		local curr = self[x]; x=x+1
		local k = ink + pap*8
		local t = curr.cache[k]
		if t then return t[1],t[2] end
		
		-- no change in ink/paper
		local c,u = 64,self:findRec(x, ink, pap)
		t = curr:calcErr(ink, pap) + u
		if ink+pap~=7 then
			u = curr:calcErr(7-ink,7-pap) + u
			if u<t then t,c = u,c+128 end
		end
			
		-- ink change
		local v,w = curr:calcErr(7-pap, 7-pap),0
		u = curr:calcErr(pap, pap)
		if v<u then u,w = v,w+128 end
		if u<t then
			for i=0,7 do
				if i~=ink then
					v = u + self:findRec(x, i, pap)
					if v<t then	t,c = v,w+i end
				end
			end
		end
			
		-- pap change
		for i=0,7 do 
			if i~=pap then
				u = self:findRec(x, ink, i)
				if u<t then
					v = curr:calcErr(  i,  i) + u; if v<t then t,c = v,    16+i end
					v = curr:calcErr(7-i,7-i) + u; if v<t then t,c = v,128+16+i end
				end 
			end
		end
		
		curr.cache[k] = {t,c}
		return t,c
	end
	
	return line
end

function optim(y, err1) 
	local line = mkLine(y,err1)
	if running>0 then
		local ink,pap = 7,0		
		for x=0,w39 do
			local i,_,c = false, line:findRec(x,ink,pap)
			line[x].cmd = c
			if c>=128 then c,i = c-128,true end
			if c==64 then
				line[x].fg = i and 7-ink or ink
				line[x].bg = i and 7-pap or pap
			else
				if c<8 then ink = c else pap = c-16 end
				line[x].fg = i and 7-pap or pap
				line[x].bg = i and 7-pap or pap
			end
		end
	end
	return line
end

local start = os.clock()
local tot = 0
local res = {}
local buf = ''
for i=1,width*height do res[i]=0 end
for y=0,height-1 do
	info(string.format("%2d%%", 100*y/height))	
	local line = oric>0 and optim(y,err1)
	local x0,x1,xs = 0,width-1,1
	if y%2==1 then x0,x1,xs = x1,x0,-xs end
	for x=x0,x1,xs do
		local p = getLinearPixel(x,y)
		local e = err1[x]
		e:add(p)
		local c
		if line then
			local cmd = line[math.floor(x/6)]
			local d1,d2 = dist2(e,pal[cmd.fg]),dist2(e,pal[cmd.bg])
			if d1<=d2 then
				c = cmd.fg
				tot = tot + d1
				if (cmd.cmd % 128)>=64 then
					cmd.cmd = cmd.cmd + 2^(5-(x%6))
				end
			else
				c = cmd.bg
				tot = tot + d2
			end
		else
			c=0
			local d=dist2(e,pal[c])
			for i=1,7 do
				local t=dist2(e,pal[i])
				if t<d then c,d = i,t end
			end
			tot = tot + d
		end
		res[y*width + x + 1] = c
		e:sub(pal[c])
		ostro:diffuse(p, e, err1[x+xs], err2[x-xs], err2[x])
	end
	err1,err2 = err2,err1
	err2:clear()
	if oric>0 then
		local lbuf = {}
		for i=0,w39 do lbuf[i+1] =line[i].cmd end
		buf = buf .. string.char(unpack(lbuf))
	end
end

info("time="..(math.floor(.5 + 100*(os.clock()-start))/100).."s err="..tot^.5)

if oric>0 then
	local function exist(file)
		local f=io.open(file,'rb')
		if not f then return false else io.close(f); return true; end
	end
	local name,path = getfilename()
	name = string.gsub(name,"%.%w*$","")
	local tapname = name .. ".tap"
	local fullname = path .. '/' .. tapname
	local ok = not exist(fullname)
	if not ok then
		selectbox("Ovr " .. tapname .. "?", "Yes", function() ok = true; end, "No", function() ok = false; end)
	end
	if ok then 
		-- http://caca.zoy.org/browser/libpipi/trunk/pipi/codec/oric.c
		local out = io.open(fullname,"wb")
		out:write(string.char(0x16,0x16,0x16,0x16,0x24,
			0x00,0xff,0x80,0x00,0xbf,0x3f,0xa0,0x00,0x00) .. 
			name .. 
			string.char(0) ..
			buf)
		out:close()
	end
end

setpicturesize(width,height)
pal:disp()
for i,v in ipairs(res) do
	putpicturepixel((i-1)%width,math.floor((i-1)/width),v)
end



-- finalizepicture()
--[[
-- file = io.open('C:/Users/Utilisateur/Downloads/einst.tap','r')
file = io.open('C:/Users/Utilisateur/Downloads/output.tap','r')

data = file:read('*a')
file:close()

s=0
i=1
t=''
while i<=data:len() and s~=11 do
	local b = data:byte(i)
	-- messagebox(s)
	if s==0 then
		s = b==0x24 and 1 or b==0x16 and s or -1
	elseif s==1 then 
		s = b==0x00 and 2 or -1
	elseif s==2 then
		s= b==0xff and 3 or -1
	elseif s==3 then
		s=b==0x80 and 4 or -1
	elseif s==4 then
		s = b==0x00 and 5 or -1
	elseif s==5 then
		s = b==0xbf and 6 or -1
	elseif s==6 then
		s = b==0x3f and 7 or -1
	elseif s==7 then
		s = b==0xa0 and 8 or -1
	elseif s==8 then
		s = b==0x00 and 9 or -1
	elseif s==9 then
		s = 10
	elseif s==10 then
		s = b==0x00 and 11 or s
	else
		error('invalid tap file at offset: ' .. i)
	end
	if (s==9 or s==10) and b~=0 then
		t = t..string.char(b)
	end
	i = i + 1
end
messagebox(t)

for y=0,199 do
	local paper,ink=0,7
	for x=0,39 do
		local b,z = data:byte(i),false i=i+1
		if b>=128 then b,z=b-128,true end
		if b<64 then
			if b<8 then
				ink = b
			elseif b>=16 and b<24 then
				paper = b-16
			end
			b = 0
		else
			b = b - 64
		end
		for x2=x*6,x*6+5 do
			if b>=32 then
				b = 2*(b-32)
				putpicturepixel(x2,y,z and 7-ink or ink)
			else
				b = 2*b
				putpicturepixel(x2,y,z and 7-paper or paper)
			end
		end
	end
end
]]--